package examples.jce.rsa_aes;

import java.io.*;
import javax.crypto.*;
import java.security.*;
import java.security.cert.*;
import javax.crypto.spec.SecretKeySpec;
import javax.crypto.spec.IvParameterSpec;

/**
 * Class housing methods to decrypt files with hmacs.
 */


public class DecryptHMAC
{
    private SecretKey    mac_key = null; // Key for MAC.
    private SecretKey    aes_key = null; // Key for AES Cipher.
    private byte[]       aes_iv = null;  // IV for AES in CBC mode.
    private byte[]       mac_code = null; // Mac code. (The output of the MAC)



    static final int            FILE_HEADER = 0x7e01; // File header.
    static final int            DATA_BLOCK = 1; // Data block header.
    static final int            FINAL_DATA_BLOCK = 2; // Marker for a final data block. (ie the last one)
    static final int            HMAC_BLOCK = 3; // Marker for the HMAC
   
    static final int            KEY_BLOCK = 16; // Marker for a key block.
    static final int            IV_BLOCK = 17; // Marker for the IV block.
    static final int            HMAC_KEY_BLOCK = 18; // Marker for HMAC block
    

    public void DecryptHMAC(){}

     
    /**
     * zero out the passed in byte array
     */
    private static void blank(
                              byte[]   bytes)
    {
        for (int t = 0; t < bytes.length; t++)
            {
                bytes[t]=0;
            }
    }
    
  

    /**
     * Method to decrypt the files.
     */
    
    public void decryptFile(
                            File    file_in,
                            File    file_out)
        throws Exception
    {
        
        //
        // Set up a cipher object for AES in CBC mode with PKCS7Padding
        // NOTE:
        // While encryption can be performed in one sequential operation
        // to decrypt and validate the file takes two passes.
        // I have chosen to add another separate pass that decrypts and parses the header of the encrypted file.
        // So the 'aes_key' and 'aes_iv' values have been set before this method gets called.
        // If not you will get a NullPointerException.
        //

        Cipher              cipher = Cipher.getInstance("AES/CBC/PKCS7Padding","BC"); // Symmetric cipher
        cipher.init(Cipher.DECRYPT_MODE, aes_key, new IvParameterSpec(aes_iv)); // In decrypt mode.
  
        FileOutputStream    file_str = new FileOutputStream(file_out); // To Save Plain text.
        DataInputStream     data_str = new DataInputStream(new FileInputStream(file_in)); // To read encrypted file.
        int                 cmd = 0; // variable to store block header type in.
        int                 l = 0; // Universal length holding variable.
        byte[]              buf = new byte[8192]; // A buffer to work in.
	byte[] 	            _out = null;

        for (;;)
        {
            cmd = data_str.readShort(); // Read block header.

            if (cmd == DATA_BLOCK) // Test for data block.
            {
                l = data_str.readInt(); // Get length
                data_str.readFully(buf,0,l); // Read all that (l) data.
        	_out = cipher.update(buf,0,l);   
		if (_out != null) file_str.write(_out); // Decrypt and save the data.
                System.out.print(".");
            }
        
            if (cmd == FINAL_DATA_BLOCK) // Test for last data block.
            {
                l = data_str.readInt(); // Get length
                data_str.readFully(buf,0,l); // Read all that (l) data.
        	_out = cipher.doFinal(buf,0,l);        
		if (_out != null) file_str.write(_out); // Do final decryption and save remaining data.
                System.out.println("!");
                break; // Stop looping at that point.
            }

            //
            // The following blocks need to be skipped.
            //

            if (cmd == HMAC_BLOCK) // Skip HMAC block.
                {
                    l = data_str.readInt();
                    data_str.skip(l);
                }

            if (cmd == KEY_BLOCK) // Skip the KEY Block
                {
                    l = data_str.readInt();
                    data_str.skip(l);
                }

            if (cmd == IV_BLOCK) // Skip the IV block.
                {
                    l = data_str.readInt();
                    data_str.skip(l);
                }
            
            if (cmd == HMAC_KEY_BLOCK) // Skip the HMAC_KEY_BLOCK
                {
                    l = data_str.readInt();
                    data_str.skip(l);
                }

        }

        blank(buf); // Erase the content of the buffer byte array.
        buf = null; // Mark it for collection.

        //
        // Flush and close output.
        //

        file_str.flush();
        file_str.close();
        data_str.close();
    }


    /**
     * Validate the encrypted file by using the MAC code.
     * Like the decrypt method this method assumes that the
     * parseHeader() method has already been called and
     * the 'mac_key' variable had been set.
     * If not it will break with a NullPointerException.
     * <P>
     * To validate (Note: not authenticate) the file a
     * the mac is first initialized with the mac key and
     * then the whole file is processed through the MacInputStream
     * when the HMAC_BLOCK is found the processing stops and
     * the mac code is computer. This computed mac code is compared
     * to the mac code stored in the HMAC_BLOCK.
     * If the two codes match then the file is valid and intact.
     */

   
    public boolean validateHMAC(
        File    file_in)
        throws Exception
    {
        Mac             mac = Mac.getInstance("HMACSHA1","BC"); // Mac algorithm based on SHA1 message digest.

        mac.init(mac_key); // initialize it with the mac key.
    
        //
        // The use of the MacInputStream filter makes processing the file a very simple process.
        // As data is read through the filter the mac 'mac' is updated.
        // When we run out of data to read we simply call the doFinal() method on the MAC
        // and get our mac code.
        //


        MacInputStream  mac_str = new MacInputStream(new FileInputStream(file_in),mac); // Wrapped input stream.
        DataInputStream data_str = new DataInputStream(mac_str); // Then wrap with a DataInputStream.
        int             cmd = 0;
        int l=0;
        byte[] buf = new byte[8192];
        do
        {
            cmd = data_str.readShort();

            if (cmd == DATA_BLOCK) // Skip DATA block.
                {
                    l = data_str.readInt();
                    data_str.read(buf,0,l); // Dummy read.
                }

            if (cmd == FINAL_DATA_BLOCK) // Skip FINAL_DATA block.
                {
                    l = data_str.readInt();
                    data_str.read(buf,0,l); // Dummy read
                }

            
            if (cmd == KEY_BLOCK) // Skip the KEY Block
                {
                    l = data_str.readInt();
                    data_str.read(buf,0,l); // Dummy read
                }

            if (cmd == IV_BLOCK) // Skip the IV block.
                {
                    l = data_str.readInt();
                    data_str.read(buf,0,l); // Dummy Read
                }
            
            if (cmd == HMAC_KEY_BLOCK) // Skip the HMAC_KEY_BLOCK
                {
                    l = data_str.readInt();
                    data_str.read(buf,0,l); // Dummy read.
                }


        }

        while (cmd != HMAC_BLOCK);
    
        //
        // You will notice that I am not reading off the mac code from the HMAC_BLOCK.
        // Well thats because the parseHeader() method has already read the offending
        // data into the 'mac_code' variable.
        //

        buf = mac.doFinal();   // Reuse buf. 
        data_str.close(); // Close input stream.

        //
        // The java.security.MessageDigest class contains a very handy little byte array
        // comparison routine that returns a boolean true when the two byte arrays match.
        // In this instate I simply return the output of that method to the caller.
        //

        return MessageDigest.isEqual(buf,mac_code);
    }


    /**
     * The parseHeader() method, parses the header and attempts to decrypt any information
     * that needs to be decrypted.
     * To do this the parse header method requires the private key of the recipient.
     *
     * In these examples RSA is used with OAEPPadding. OAEPPadding is very very useful.
     * The encoding is such that if the decryption fails the decoding of the OAEPPadding
     * will fail also. (See RSAES-OAEP over at RSA labs.)
     */


    public void parseHeader(
        File            file_in,
        PrivateKey      priv_key,
        SecureRandom    sec_rand)
        throws Exception
    {
        DataInputStream data_in = new DataInputStream(new FileInputStream(file_in)); // InputStream for the file.
        int             l = 0; // Universal length variable.
        boolean         ena = false; // A flag to determine if the file header has been read.
        boolean         stop = false; // A flag to stop.
        Cipher          rsa_eng = Cipher.getInstance("RSA/None/OAEPPadding", "BC"); // RSA cipher.
        rsa_eng.init(Cipher.DECRYPT_MODE,priv_key,sec_rand); // Set up Cipher object with private key.

        while(!stop) // Loop
        {           
            try
            {
                int cmd = data_in.readShort(); // Read in block header.
                
                if (cmd == FILE_HEADER) // A file header.
                {
                    ena = true; // Set true. File header has been found!
                    System.out.println("Header Parse: File Header");
                    continue;
                }
                
                if (cmd == DATA_BLOCK) // Data block. (The content is skipped in this method)
                {
                    if (!ena)
                    {
                        throw new Exception("Broken header");
                    }

                    System.out.println("Header Parse: Data Block, size = "+l);
                    l = data_in.readInt(); // Get length
                    data_in.skip(l);  // Skip.
                    continue; 
                }
                
                if (cmd == FINAL_DATA_BLOCK) // Final data block
                {
                    if (!ena)
                    {
                        throw new Exception("Broken header");
                    }

                    l = data_in.readInt(); // Read length
                    System.out.println("Header Parse: Final Data Block, size = "+l);
                    data_in.skip(l);  // Skip data.
                    continue;
                }
                
                if (cmd == HMAC_BLOCK) // mac code block.
                {
                    if (!ena)
                    {
                        throw new Exception("Broken header");
                    }

                    l = data_in.readInt(); // Read length.
                    System.out.println("Parse: HMAC block size = "+l);
                    mac_code = new byte[l]; // Create new array.
                    data_in.readFully(mac_code); // Read in data.
                    continue;
                }
                
                if (cmd == KEY_BLOCK) // Key block
                {
                    if (!ena)
                    {
                        throw new Exception("Broken header");
                    }

                    l = data_in.readInt(); // Read length
                    System.out.println("Parse: Key encoded block size = "+l);

                    byte[] d = new byte[l]; // Create new array.

                    data_in.readFully(d); // Read in raw AES key.

                    //
                    // At this stage the raw data needs to be converted back into
                    // a SecretKey. We use SecretKeySpec to do this.
                    // See java doc on SecretKeySpec for more details.
                    // Note: I have also tucked away an rsa_eng.doFinal() to
                    // decrypt the data before conversion.
                    //

                    aes_key = (SecretKey)new SecretKeySpec(rsa_eng.doFinal(d), "AES"); // Decrypt and Convert.
                    continue;
                }
                
                if (cmd == IV_BLOCK) // Iv block.
                {
                    if (!ena)
                    {
                        throw new Exception("Broken header");
                    }

                    l = data_in.readInt();// Read length
                    System.out.println("Parse: IV block, size = " + l);
                    aes_iv = new byte[l]; // Create new byte array.
                    data_in.readFully(aes_iv); // Read in data.
                    aes_iv = rsa_eng.doFinal(aes_iv); // Decrypt.
                    continue;
                }
                

                if (cmd == HMAC_KEY_BLOCK)
                {
                    if (!ena)
                    {
                        throw new Exception("Broken header");
                    }

                    l = data_in.readInt(); // Read length
                    System.out.println("Parse: HMAC Key block, size = "+l);
                    byte[] d = new byte[l]; // Create new Array.

                    data_in.readFully(d); // Read in data.

                    //
                    // At this stage the raw data needs to be converted back into
                    // a SecretKey. We use SecretKeySpec to do this.
                    // See java doc on SecretKeySpec for more details.
                    // I have also tucked away an rsa_eng.doFinal() to
                    // decrypt the data before conversion.
                    // You will notice that the name of the cipher specified is AES
                    // yet we are building a key for HMACSHA1. This is because at the
                    // moment the BouncyCastle implementation of JCE does not recognize
                    // HMACSHA1. This will be fixed shortly. However is it safe to use
                    // 'AES' because the AES conversion just takes the data directly
                    // and converts it to a secret key.
                    //

                    //
                    // Notice that the HMAC key is encrypted with the Symmetric cipher.
                    // This is done because an attacker could foil the system by replacing
                    // this block with something they generated if they know the recipients
                    // public key.
                    //
                    // The attacker does not know the AES key or IV so in this instance
                    // the attacker can not easily modify the file without first figuring out
                    // what the AES key is..
                    // 
                    // However because there is no authentication the attacker could just create
                    // a whole new file instead..
                    //

                    Cipher hmac_dec = Cipher.getInstance("AES/CBC/PKCS7Padding","BC");
                    hmac_dec.init(Cipher.DECRYPT_MODE,aes_key,new IvParameterSpec(aes_iv));
                    mac_key = (SecretKey)new SecretKeySpec(hmac_dec.doFinal(d), "HMACSHA1"); // Read and convert.
                    continue;
                }



            }
            catch (EOFException eof)
            {
                stop = true;
            }
        }

        return;
    }







}
