package examples.jce.rsa_aes;

import java.io.*;
import javax.crypto.*;
import javax.crypto.spec.*;
import java.security.*;
import java.security.cert.*;



/**
 * Encrypt a file using HMAC for message integrity checking.
 * This class can be used from the command line
 * EncryptFileHMAC [source file] [dest file] [recipients certificate]
 */


public class EncryptFileHMAC
{
    private File file_in = null; // Source File.
    private File file_out = null; // Output File.
    private SecureRandom sec_rand = null; // Secure Random instance. 
    private PublicKey  dest_key = null; // Recipients public key.




    /**
     * Main Method.
     */

    public static void main(String[] args)
    {
        new EncryptFileHMAC(args);
    }


    /**
     * Blank constructor to enable class creation without having to provide parameters
     */

    public EncryptFileHMAC()
    {
        // Blank constructor.
    }


    /**
     * Constructor designed more for command line processing
     */

    public EncryptFileHMAC(String[] args)
    {
        try {
            initTerm(args);
            encryptFile();
        } catch (Exception ex)
            {
                ex.printStackTrace();
            }
    }
     


    /**
     * Construct and initialise
     */

    public EncryptFileHMAC(File file_in, File file_out, PublicKey dest_key) throws Exception
    {
        init(file_in,file_out,dest_key);
    }



    /**
     * Iniatialise
     */

    public void init(File file_in,File file_out,PublicKey dest_key) throws Exception
    {
   
        Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider());
        sec_rand = SecureRandom.getInstance("SHA1PRNG","SUN");
        
        if (new File("/dev/urandom").exists())
            {
                System.out.println("Salting SecureRandom (SHA1PRNG) from /dev/urandom");
                byte[] salt = new byte[8192];
                new FileInputStream("/dev/urandom").read(salt);
                sec_rand.setSeed(salt);
            }
   

        this.file_in = file_in;
        this.file_out = file_out;
        this.dest_key = dest_key;
    }



    /**
     * Initialise from terminal and command line inputs
     */

    public void initTerm(String[] args)
    {
          try
        {
            Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider());
            sec_rand = SecureRandom.getInstance("SHA1PRNG","SUN");
            
            if (new File("/dev/urandom").exists())
            {
                System.out.println("Salting SecureRandom (SHA1PRNG) from /dev/urandom");
                byte[] salt = new byte[8192];
                new FileInputStream("/dev/urandom").read(salt);
                sec_rand.setSeed(salt);
            }
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }

        if (args.length <3)
        {
            System.out.println("Usage: <source file> <dest file> <recipients certificate>");
            System.exit(0);
        }

        file_in = new File(args[0]);
        file_out = new File(args[1]);

        if (!file_in.exists())
        {
            System.out.println("Source File does not exist..!");
            System.exit(0);
        }
    
        File cert_file = new File(args[2]);

        if (!cert_file.exists())
        {
            System.out.println("Recipients Certificate File does not exist..!");
            System.exit(0);
        }

        //
        // Read in certificate.
        // See java doc for CertificateFactory for more info..
        //

        System.out.println("Reading recipients certificate file.");

        try
        {
            byte[]                  raw = new byte[(int)cert_file.length()];
            DataInputStream         din = new DataInputStream(new FileInputStream(cert_file));

            din.readFully(raw);
            din.close();
    
            ByteArrayInputStream    b_in = new ByteArrayInputStream(raw);
            CertificateFactory      cf = CertificateFactory.getInstance("X.509");
            X509Certificate         x_cert = (X509Certificate)cf.generateCertificate(b_in);
            dest_key = x_cert.getPublicKey();

        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }



    /**
     * Encrypt file method
     */

    public void encryptFile() throws Exception
    {
        EncryptHMAC.encryptFile(file_in, dest_key, sec_rand, file_out);
    }





    /**
     * Read line in from System.in
     */

    private String getLine()
    {
        try
        {
            BufferedReader bin = new BufferedReader(new InputStreamReader(System.in));
            return(bin.readLine());
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }

        return(null);
    }



  

 
      
}




