package examples.jce.session;

import java.io.*;
import javax.crypto.*;
import java.security.*;


/**
 * This is an extended version of DataInputStream
 * By extending DataInputStream we can create some methods
 * that make our life easier.
 * The readsByteArray() the length of the array then the
 * array content.
 * The readPDU reads in an encrypted block of data and
 * checks that the MAC code matched.
 * The readPDU will throw IOExceptions if then decryption
 * or MAC codes fail.
 */


public class ExDataInputStream extends DataInputStream
{
    private Cipher cipher = null;
    private Mac mac = null;
    

    public static boolean isEqual(
        byte[] a,
        byte[] b)
    {
        if (a == null || b == null)
        {
            return(false);
        }

        if (a.length != b.length)
        {
            return(false);
        }

        for (int t=0;t<a.length; t++)
        {
            if (a[t] != b[t])
            {
                return(false);
            }
        }

        return(true);
    }

    public ExDataInputStream(
        InputStream in)
    {
        super(in);
    }

    public byte[] readByteArray()
        throws IOException
    {
        byte[] out = new byte[readInt()];
        readFully(out);
        return(out);
    }

    public void initPDU(
        Cipher  cipher,
        Mac     mac)
    {
        this.cipher = cipher;
        this.mac = mac;
    }

    public byte[] readPDU()
        throws IOException
    {
        byte[] b = readByteArray(); // Read in encrypted data.
           
        try
        {
            b = cipher.doFinal(b);
        }
        catch (Exception ex)
	    {
            throw new IOException("PDU failed to decrypt "+ex.getMessage());
	    }

        mac.update(b);	// MAC is taken on plain text.

        byte[] m = null;
        m = mac.doFinal();
        if (!isEqual(m, readByteArray()))
        {
            throw new IOException("PDU Mac code failed ");
        }
       
        //
        // Load mac with previous MAC value.
        // This forces each exchange into a chain
        // so that if any of the blocks are replayed out
        // of sequence the replayed blocks will fail.
        //
        
        mac.update(m);
        
        return(b);
    }
}


