package examples.jce.session;

import java.io.*;

import javax.crypto.*;

/**
 * This is an extended version of DataOutputStream
 * By extending data OutputStream we can create some methods
 * that make our life easier.
 * The writeByteArray() writes both the length of the array then the
 * array content.
 * The writePDU method takes a byte array of plain text and creates
 * an encrypted data block with a mac attached to the end of it.
 */ 
public class ExDataOutputStream extends DataOutputStream
{
    private Cipher cipher = null;
    private Mac mac = null;
   
    public ExDataOutputStream(
        OutputStream out)
    {
        super(out);
    }


    public void writeByteArray(
        byte[] b)
    throws IOException
    {
        writeInt(b.length);
        write(b);
    }

    public void initPDU(
        Cipher cipher,
        Mac mac)
    {
        this.cipher = cipher;
        this.mac = mac;
    }

    public void writePDU(
        byte[] b)
    throws IOException
    {
        mac.update(b); // Update plain text into MAC
       
        try
        {
            b = cipher.doFinal(b); // Encrypt data for recipient.
        }
        catch (Exception bex)
        {
            throw (new IOException("PDU failed to encrypt " + bex.getMessage()));
        }
    
        writeByteArray(b); // Write out data.
        b = mac.doFinal();
        writeByteArray(b); // Write out mac
    
        //
        // Load mac with previous MAC value.
        // This forces each exchange into a chain
        // so that if any of the blocks are replayed out
        // of sequence the replayed blocks will fail.
        //
        
        mac.update(b);
    }
}
